/* See the end of this file for copyright and license terms. */

#include <gay/types.h>

#include <limits.h>

#define	wsize	sizeof(unsigned long)
#define	wmask	(wsize - 1)

/* XXX get this mess sorted out */

#ifdef BZERO
#include <strings.h>

#define	RETURN	return
#define	VAL	0
#define	WIDEVAL	0

void bzero(void *dst0, usize length)
#else /* not BZERO */
#include <string.h>

#define	RETURN	return (dst0)
#define	VAL	c0
#define	WIDEVAL	c

#ifdef __HAVE_ARCH_MEMSET
void *__memset(void *dst0, int c0, usize length)
#else
void *memset(void *dst0, int c0, usize length)
#endif
#endif /* not BZERO */
{
	usize t;
#ifndef BZERO
	unsigned long c;
#endif
	unsigned char *dst;

	dst = dst0;
	/*
	 * If not enough words, just fill bytes.  A length >= 2 words
	 * guarantees that at least one of them is `complete' after
	 * any necessary alignment.  For instance:
	 *
	 *	|-----------|-----------|-----------|
	 *	|00|01|02|03|04|05|06|07|08|09|0A|00|
	 *	          ^---------------------^
	 *		 dst		 dst+length-1
	 *
	 * but we use a minimum of 3 here since the overhead of the code
	 * to do word writes is substantial.
	 *
	 * TODO: This threshold might not be sensible for 64-bit u_long.
	 * We should benchmark and revisit this decision.
	 */
	if (length < 3 * wsize) {
		while (length != 0) {
			*dst++ = VAL;
			--length;
		}
		RETURN;
	}

#ifndef BZERO
	if ((c = (unsigned char)c0) != 0) {	/* Fill the word. */
		c = (c << 8) | c;	/* u_long is 16 bits. */
#if ULONG_MAX > 0xffff
		c = (c << 16) | c;	/* u_long is 32 bits. */
#endif
#if ULONG_MAX > 0xffffffff
		c = (c << 32) | c;	/* u_long is 64 bits. */
#endif
	}
#endif
	/* Align destination by filling in bytes. */
	if ((t = (long)dst & wmask) != 0) {
		t = wsize - t;
		length -= t;
		do {
			*dst++ = VAL;
		} while (--t != 0);
	}

	/* Fill words.  Length was >= 2*words so we know t >= 1 here. */
	t = length / wsize;
	do {
		*(unsigned long *)(void *)dst = WIDEVAL;
		dst += wsize;
	} while (--t != 0);

	/* Mop up trailing bytes, if any. */
	t = length & wmask;
	if (t != 0) {
		do {
			*dst++ = VAL;
		} while (--t != 0);
	}
	RETURN;
}

#ifndef BZERO
#ifndef __HAVE_ARCH_MEMSET16
void *memset16(u16 *dest, u16 val, usize nbyte)
{
	void *dst0 = dest;

	if (nbyte >= 2) {
		if ((uintptr_t)dest % 4)
			*dest++ = val;
		nbyte -= 2;
		if (nbyte)
			memset32((u32 *)dest, val | ((u32)val << 16), nbyte);
	}

	return dst0;
}
#endif /* !__HAVE_ARCH_MEMSET16 */

#ifndef __HAVE_ARCH_MEMSET32
void memset32(u32 *dest, u32 val, usize nbyte)
{
	void *dst0 = dest;

	if (nbyte >= 4) {
#if LONG_BIT >= 64
		if ((uintptr_t)dest % 8)
			*dest++ = val;
		nbyte -= 4;
		if (nbyte >= 8)
			return memset64((u64 *)dest, val | ((u64)val << 32), nbyte);
#else
		usize n = nbyte / 4;
		while (n--)
			*dest++ = val;
#endif
	}

	return dst0;
}
#endif /* !__HAVE_ARCH_MEMSET32 */

#ifndef __HAVE_ARCH_MEMSET64
void memset64(u64 *dest, u64 val, usize nbyte)
{
	void *dst0 = dest;

	usize n = nbyte / 8;
	while (n--)
		*dest++ = val;

	return dst0;
}
#endif /* !__HAVE_ARCH_MEMSET64 */
#endif

/*
 * Copyright (c) 2021 fef <owo@fef.moe>
 * Copyright (c) 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Hibler and Chris Torek.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */
